from ..config import *


class SceneManager:
    def __init__(self, game):
        self.game = game

        self.scenes = {}
        self.selected_scene = None

    def add_scene(self, name, scene):
        if len(self.scenes) == 0:
            self.selected_scene = name

        scene.game = self.game
        scene.name = name
        scene.display_surface = pg.display.get_surface()
        scene.on_create()

        self.scenes[name] = scene

    def event_handler(self,event):
        if self.selected_scene is not None:
            self.scenes[self.selected_scene].on_event(event)

    def render(self):
        if self.selected_scene is not None:
            self.scenes[self.selected_scene].on_render()

    def update(self):
        if self.selected_scene is not None:
            self.scenes[self.selected_scene].on_update()

    def close(self):
        if self.selected_scene is not None:
            self.scenes[self.selected_scene].on_close()



from ..config import *
from .base_component import BaseComponent,ComponentType


class RendererComponent(BaseComponent):
    def __init__(self, transform_component, image: pg.Surface):

        super().__init__()
        self.type = ComponentType.RenderComponent

        self.transform = transform_component
        self.image = image





import os
import msgpack
import pygame as pg
from io import BytesIO

class AssetManager:
    def __init__(self, game):
        self.game = game
        self.root = os.path.join(self.game.build_dir, "assets.aab")
        self.scene_assets = {}

        self._init_aab_file()

    @property
    def selected_scene(self):
        return self.game.scene_manager.selected_scene or "all-scenes"

    def _init_aab_file(self):
        if not os.path.exists(self.root):
            with open(self.root, "wb") as f:
                f.write(self._encode(msgpack.packb({})))

    def _encode(self, data: bytes) -> bytes:
        return self.game.cipher.encrypt(data) if self.game.cipher else data

    def _decode(self, data: bytes) -> dict:
        decrypted = self.game.cipher.decrypt(data) if self.game.cipher else data
        return msgpack.unpackb(decrypted)

    def _load_assets(self) -> dict:
        with open(self.root, "rb") as file:
            return self._decode(file.read())

    def _save_assets(self, assets: dict):
        packed = msgpack.packb(assets, use_bin_type=True)
        with open(self.root, "wb") as file:
            file.write(self._encode(packed))

    def _surface_to_png_bytes(self, surface):
        buffer = BytesIO()
        pg.image.save(surface, buffer, "PNG")
        return buffer.getvalue()

    def save_image(self, scene_name, path):
        image = pg.image.load(path)
        self.scene_assets.setdefault(scene_name, {})[path] = image

        image_data = {
            "size": image.get_size(),
            "raw-data": self._surface_to_png_bytes(image)
        }

        assets = self._load_assets()
        assets.setdefault(scene_name, {})[path] = image_data
        self._save_assets(assets)

        return image

    def load_image(self, scene_name, path):
        try:
            assets = self._load_assets()
            if scene_name in assets and path in assets[scene_name]:
                image_data = assets[scene_name][path]
                surface = pg.image.load(BytesIO(image_data["raw-data"]))
                self.scene_assets.setdefault(scene_name, {})[path] = surface
                return surface
        except Exception as e:
            print(f"[IMAGE-LOAD] Error loading from `.aab`: {e}")

        return self.save_image(scene_name, path)

    def get_image(self, scene_name, path):
        if scene_name in self.scene_assets and path in self.scene_assets[scene_name]:
            return self.scene_assets[scene_name][path]

        self.scene_assets.setdefault(scene_name, {})
        return self.load_image(scene_name, path)

    def save_sound(self, scene_name, path):
        sound = pg.mixer.Sound(path)
        self.scene_assets.setdefault(scene_name, {})[path] = sound

        with open(path, "rb") as file:
            raw_data = file.read()

        sound_data = {
            "raw-data": raw_data
        }

        assets = self._load_assets()
        assets.setdefault(scene_name, {})[path] = sound_data
        self._save_assets(assets)

        return sound

    def load_sound(self, scene_name, path):
        try:
            assets = self._load_assets()
            if scene_name in assets and path in assets[scene_name]:
                raw = assets[scene_name][path]["raw-data"]
                sound = pg.mixer.Sound(file=BytesIO(raw))
                self.scene_assets.setdefault(scene_name, {})[path] = sound
                return sound
        except Exception as e:
            print(f"[SOUND-LOAD] Error loading from `.aab`: {e}")

        return self.save_sound(scene_name, path)

    def get_sound(self, scene_name, path):
        if scene_name in self.scene_assets and path in self.scene_assets[scene_name]:
            return self.scene_assets[scene_name][path]

        self.scene_assets.setdefault(scene_name, {})
        return self.load_sound(scene_name, path)



from ..components.base_component import ComponentType
from ..components.renderer_component import RendererComponent
from ..config import *

class Renderer:
    def __init__(self,render_target=None):
        self.game = None
        self.scene = None

        self.render_target = render_target

        self.render_objects = []

    def check_dirty(self):
        if self.scene.dirty:
            self.render_objects = []
            for object in self.scene.game_objects:
                if render_comp := object.get_component(ComponentType.RenderComponent):
                    self.render_objects.append(render_comp)

            self.scene.dirty = False

    def create(self):
        self.render_target = self.render_target if not self.render_target is None else self.scene.display_surface
        self.check_dirty()

    def update(self):
        self.check_dirty()

    def render(self):
        for render_comp in self.render_objects:
            self.render_target.blit(render_comp.image,render_comp.transform.rect)


from ..components.base_component import ComponentType
from ..config import *

class BaseGameObject:
    def __init__(self):
        self.game = None
        self.scene = None
        self.id = None

        self.dirty = False

        self.components = {}

    def add_component(self,component):
        if component.type is None:
            raise ValueError(f"Component {component} is missing a valid 'type'. Cannot add to GameObject.")

        component.game = self.game
        component.scene = self.scene
        component.game_object = self

        self.dirty = True

        self.components[component.type] = component

    def get_component(self,component_type:ComponentType):
        if component_type in self.components:
            return self.components[component_type]
        return None

    def on_create(self): ...

    def on_update(self):
        for component in self.components.values():
            component.on_update()

    def on_close(self):
        for component in self.components.values():
            component.on_close()

            self.game = None
            self.scene = None
            self.components = []


from .components.transform_component import TransformComponent
from .components.renderer_component import RendererComponent

from .components.base_component import (
BaseComponent,
ComponentType
)

from .game_object.base_object import (
BaseGameObject
)

from .scenes.base_scene import (
BaseScene
)

from .core.window import (
WINDOW_ICON,
WINDOW_TITLE,
WINDOW_POS,
WINDOW_BG_COLOUR,
WINDOW_SIZE,
window_config
)

from .game import (
Game,
GAME_BUILD_LOCATION,
GAME_FPS_LOCk,
GAME_FILE_ENCRYPTION_KEY,
game_config
)


from ..config import *
from enum import Enum,auto

class ComponentType(Enum):
    RenderComponent = auto()
    TransformComponent = auto()

class BaseComponent:
    def __init__(self):
        self.game = None
        self.scene = None
        self.game_object = None

        self.type = None

    def on_update(self): ...

    def on_close(self): ...


import os.path

from .config import *
from .core.window import Window
from .core.scene_manager import SceneManager
from .core.asset_manager import AssetManager

GAME_FPS_LOCk = "game-fps-lock"
GAME_BUILD_LOCATION = "game-build-location"
GAME_FILE_ENCRYPTION_KEY = "game-file-encryption-key"

game_config = {
    GAME_FPS_LOCk: 0,
    GAME_BUILD_LOCATION: os.getcwd(),
    GAME_FILE_ENCRYPTION_KEY: None,
}

class Game:
    def __init__(self,extra_info=False):
        pg.init()

        self.extra_info = extra_info

        self.build_dir = game_config[GAME_BUILD_LOCATION]
        self.encrypt_key = game_config[GAME_FILE_ENCRYPTION_KEY]
        self.cipher = Fernet(self.encrypt_key) if self.encrypt_key else None

        self.root = os.path.abspath(os.path.dirname(__file__))
        print(self.root)

        self.asset_manager = AssetManager(self)
        self.window = Window(self)

        self.scene_manager = SceneManager(self)

        self.clock = pg.time.Clock()

        self.dt = 1

        self.running = True

    def event_handler(self):
        for event in pg.event.get():
            self.window.event_handler(event)

            self.scene_manager.event_handler(event)

    def render(self):
        self.window.render()

        self.scene_manager.render()

        pg.display.flip()

    def update(self):
        self.dt = self.clock.tick(game_config[GAME_FPS_LOCk]) / 1000

        self.scene_manager.update()

        self.window.update()

    def close(self):
        self.scene_manager.close()
        self.window.close()
        pg.quit()
        sys.exit()

    def run(self):
        while self.running:
            self.event_handler()
            self.update()
            self.render()

        self.close()


from ..config import *
from ..core.renderer import Renderer

class BaseScene:
    def __init__(self):
        self.game = None
        self.name = None
        self.display_surface = None

        self.dirty = False

        self.renderer = Renderer()


        self.game_objects = []

    def get_image(self,path,*args,**kwargs):
        if not self.game:
            raise RuntimeError("\n\nDo not call 'get_image()' in '__init__'. Wait until the scene is fully initialized (e.g., in 'on_create()').")

        return self.game.asset_manager.get_image(kwargs.get("all-scenes",self.name), path)

    def get_sound(self,path,*args,**kwargs):
        if not self.game:
            raise RuntimeError(
                "\n\nDo not call 'get_sound()' in '__init__'. Wait until the scene is fully initialized (e.g., in 'on_create()').")

        return self.game.asset_manager.get_sound(kwargs.get("all-scenes", self.name), path)

    def add_game_object(self,obj):
        obj.id = len(self.game_objects)
        obj.scene = self
        obj.game = self.game
        obj.on_create()

        self.dirty = True

        self.game_objects.append(obj)

    def remove_game_object(self, obj):
        if obj in self.game_objects and obj.id != -1:
            self.game_objects.remove(obj)
            obj.id = -1
            obj.on_close()
            self.dirty = True
            # Recalculate IDs
            for index, game_obj in enumerate(self.game_objects):
                game_obj.id = index

    def on_create(self):
        self.renderer.game = self.game
        self.renderer.scene = self
        self.renderer.create()

    def on_event(self,event):
        pass

    def on_update(self):
        for game_obj in self.game_objects:
            game_obj.on_update()
        self.renderer.update()

    def on_render(self):
        self.renderer.render()

    def on_close(self):
        pass


from ..config import *

WINDOW_SIZE = "window-size"
WINDOW_POS = "window-pos"
WINDOW_TITLE = "window-title"
WINDOW_BG_COLOUR = "window-bg-colour"
WINDOW_ICON = "window-icon"

window_config = {
    WINDOW_SIZE: (1080, 720),
    WINDOW_POS: (-1, -1),
    WINDOW_TITLE: "========== Default Title ==========",
    WINDOW_BG_COLOUR: (10,40,80),
    WINDOW_ICON: None,
}

class Window:
    def __init__(self,game):
        self.game = game

        icon_path = window_config[WINDOW_ICON] if window_config[WINDOW_ICON] else os.path.join(self.game.root,"assets", "Icon.png")

        self.icon = self.game.asset_manager.get_image("all-scenes",icon_path)
        pg.display.set_icon(self.icon)

        self.win = pg.display.set_mode(window_config[WINDOW_SIZE])
        pg.display.set_caption(window_config[WINDOW_TITLE])

        screen_w, screen_h = pg.display.get_desktop_sizes()[0]
        win_w, win_h = self.win.get_size()

        # Calculate centered position
        center_x = (screen_w - win_w) // 2
        center_y = (screen_h - win_h) // 2

        pos = [ window_config[WINDOW_POS][0], window_config[WINDOW_POS][1]]

        if window_config[WINDOW_POS][0] == -1:
            pos[0] = center_x

        if window_config[WINDOW_POS][1] == -1:
            pos[1] = center_y

        self.width,self.height = self.win.get_size()

        pg.display.set_window_position(pos)

    def event_handler(self,event):
        if event.type == pg.QUIT:
            self.game.close()

    def render(self):
        self.win.fill(window_config[WINDOW_BG_COLOUR])

    def update(self):
        if self.game.extra_info:
            pg.display.set_caption(f"{window_config[WINDOW_TITLE]} | FPS: {self.game.clock.get_fps() :.0f}")

    def close(self):
        pass




from ..config import *
from .base_component import BaseComponent,ComponentType

class TransformComponent(BaseComponent):
    def __init__(self,position:type[int,int]=(0,0),
                 rotation:int=0,
                 size:type[int,int]=(100,100)):
        super().__init__()

        self.type = ComponentType.TransformComponent

        self.position = position
        self.rotation = rotation
        self.size = size

        self.rect = pg.Rect(self.position,self.size)


from setuptools import setup, find_packages

setup(
    name="pygame_framework_1just_josh",
    version="0.0.9",
    author="Joshua - jjboy2245",
    description="this is a simple pygame wrapper for advanced game features with less boilerplate",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    license="Apache-2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
    ],
    python_requires=">=3.7",
    install_requires=[
        "pygame-ce",
        "msgpack",
        "cryptography",
    ],
)


import os
import sys
import msgpack

os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"

import pygame as pg

from pathlib import Path
from io import BytesIO
from cryptography.fernet import Fernet

def is_valid_file_path(path_str):
    path = Path(path_str)
    return path.is_file()

